/*
 * Copyright (C) 2017 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License
 */

package android.support.tools.jetifier.standalone

import android.support.tools.jetifier.core.FileMapping
import android.support.tools.jetifier.core.Processor
import android.support.tools.jetifier.core.config.Config
import android.support.tools.jetifier.core.config.ConfigParser
import android.support.tools.jetifier.core.utils.Log
import org.apache.commons.cli.CommandLine
import org.apache.commons.cli.DefaultParser
import org.apache.commons.cli.HelpFormatter
import org.apache.commons.cli.Option
import org.apache.commons.cli.Options
import org.apache.commons.cli.ParseException
import java.io.File
import java.nio.file.Paths

class Main {

    companion object {
        const val TAG = "Main"
        const val TOOL_NAME = "standalone"

        val OPTIONS = Options()
        val OPTION_INPUT = createOption("i", "Input libraries paths", multiple = true)
        val OPTION_OUTPUT_DIR = createOption("outputdir", "Output directory path",
                isRequired = false)
        val OPTION_OUTPUT_FILE = createOption("outputfile", "Output file", isRequired = false)
        val OPTION_CONFIG = createOption("c", "Input config path", isRequired = false)
        val OPTION_LOG_LEVEL = createOption("l", "Logging level. debug, verbose, error, info " +
            "(default)", isRequired = false)
        val OPTION_REVERSED = createOption("r", "Run reversed process", hasArgs = false,
            isRequired = false)
        val OPTION_REWRITE_SUPPORT_LIB = createOption("s", "If set, all libraries being rewritten" +
            " are assumed to be part of Support Library. Otherwise only general dependencies are" +
            " expected.",
            hasArgs = false, isRequired = false)

        private fun createOption(
            argName: String,
            desc: String,
            hasArgs: Boolean = true,
            isRequired: Boolean = true,
            multiple: Boolean = false
        ): Option {
            val op = Option(argName, hasArgs, desc)
            op.isRequired = isRequired
            if (multiple) {
                op.args = Option.UNLIMITED_VALUES
            }
            OPTIONS.addOption(op)
            return op
        }
    }

    fun run(args: Array<String>) {
        val cmd = parseCmdLine(args)
        if (cmd == null) {
            System.exit(1)
            return
        }

        Log.setLevel(cmd.getOptionValue(OPTION_LOG_LEVEL.opt))

        val inputLibraries = cmd.getOptionValues(OPTION_INPUT.opt).map { File(it) }.toSet()
        val outputDir = cmd.getOptionValue(OPTION_OUTPUT_DIR.opt)
        val outputFile = cmd.getOptionValue(OPTION_OUTPUT_FILE.opt)
        if (outputDir == null && outputFile == null) {
            throw IllegalArgumentException("Must specify -outputdir or -outputfile")
        }
        if (outputDir != null && outputFile != null) {
            throw IllegalArgumentException("Cannot specify both -outputdir and -outputfile")
        }
        if (inputLibraries.size > 1 && outputFile != null) {
            throw IllegalArgumentException(
                    "Cannot specify -outputfile when multiple input libraries are given")
        }

        val fileMappings = mutableSetOf<FileMapping>()
        if (outputFile != null) {
            fileMappings.add(FileMapping(inputLibraries.first(),
                    File(outputFile)))
        } else {
            inputLibraries.forEach {
                val newFileName = File(Paths.get(outputDir).toString(), it.name)
                fileMappings.add(FileMapping(it, newFileName))
            }
        }

        val config: Config?
        if (cmd.hasOption(OPTION_CONFIG.opt)) {
            val configPath = Paths.get(cmd.getOptionValue(OPTION_CONFIG.opt))
            config = ConfigParser.loadFromFile(configPath)
        } else {
            config = ConfigParser.loadDefaultConfig()
        }

        if (config == null) {
            Log.e(TAG, "Failed to load the config file")
            System.exit(1)
            return
        }

        val isReversed = cmd.hasOption(OPTION_REVERSED.opt)
        val rewriteSupportLib = cmd.hasOption(OPTION_REWRITE_SUPPORT_LIB.opt)
        val processor = Processor.createProcessor(
            config = config,
            reversedMode = isReversed,
            rewritingSupportLib = rewriteSupportLib)
        processor.transform(fileMappings)
    }

    private fun parseCmdLine(args: Array<String>): CommandLine? {
        try {
            return DefaultParser().parse(OPTIONS, args)
        } catch (e: ParseException) {
            Log.e(TAG, e.message.orEmpty())
            HelpFormatter().printHelp(TOOL_NAME, OPTIONS)
        }
        return null
    }
}

fun main(args: Array<String>) {
    Main().run(args)
}